import json
from pathlib import Path
import subprocess
from typing import Annotated
from fastapi import FastAPI
import typer
import sys
import getpass

from nephyx.manage.postgres import setup_database


def build_cli(fastapi_app: FastAPI, get_settings) -> typer.Typer:

    app = typer.Typer()

    @app.command()  # TODO get from settings
    def init_db():
        db_host = "localhost"
        db_port = "5432"

        db_admin_user = typer.prompt("Enter database admin user")
        db_admin_password = getpass.getpass("Enter database admin password")

        # take from config
        db_name = typer.prompt("Enter database name")
        db_user = typer.prompt("Enter database user")
        db_password = getpass.getpass("Enter database password")

        admin_config = {
            "host": db_host,
            "port": db_port,
            "user": db_admin_user,
            "password": db_admin_password,
        }

        setup_database(admin_config, db_name, db_user, db_password)

    @app.command()
    def export_openapi():
        sys.path.insert(0, "")
        openapi = fastapi_app.openapi()
        with Path("openapi.json").open("w") as f:
            json.dump(openapi, f, indent=2)

    @app.command()
    def makemigrations(comment: Annotated[str, typer.Argument()] = "auto"):
        """
        Make Alembic migrations
        """
        try:
            revision_command = f"alembic revision --autogenerate -m {comment}"
            print(f"Running Alembic migrations: {revision_command}")
            subprocess.run(revision_command, shell=True, check=True)
        except subprocess.CalledProcessError as e:
            print(f"[red]Error:[/red] {e}")
            return
        print("[green]Make migrations complete[/green]")


    @app.command()
    def migrate():
        """
        Run Alembic migrations
        """
        try:
            upgrade_command = "alembic upgrade head"
            print(f"Running Alembic upgrade: {upgrade_command}")
            subprocess.run(upgrade_command, shell=True, check=True)
        except subprocess.CalledProcessError as e:
            print(f"[red]Error:[/red] {e}")
            return
        print("[green]Migration complete[/green]")


    @app.command()
    def runserver():
        """
        Run the FastAPI server
        """
        try:
            server_command = (
                "fastapi dev app/main.py"
                if get_settings().debug
                else "fastapi run app/main.py"
            )
            print(f"Running FastAPI server: {server_command}")
            subprocess.run(server_command, shell=True, check=True)
        except subprocess.CalledProcessError as e:
            print(f"[red]Error:[/red] {e}")
            return

    return app



from collections.abc import Sequence
from typing import Any, Generic, TypeVar
from uuid import UUID

from sqlalchemy import Select, func, select
from sqlalchemy.orm import Session

#from src.domain.shared.filters import PageParams
from .model import ModelBase

ModelType = TypeVar("ModelType", bound=ModelBase)
#UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)

class BaseRepository(Generic[ModelType]):
    def __init__(self, model: type[ModelType], session: Session):
        self.model = model
        self.session = session

    def get(self, id: UUID | str) -> ModelType | None:
        return self.session.get(self.model, id)

    def get_by(self, **filters) -> ModelType | None:
        query = select(self.model)
        query = self.__filter(query, filters)
        return self.session.execute(query).scalars().first()

    def list_all(self, query_params: dict[str, Any] | None = None) -> tuple[Sequence[ModelType], int]:
        query = select(self.model)
        page_params = None
        if query_params:
            query = self.__filter(query, query_params.get("where", {}))
            query = self.__order_by(query, query_params.get("order_by", []))
            page_params = query_params.get("page_params", None)
        return self.paginate(query, page_params=page_params)


    def paginate(
            self, query: Select, *, scalars: bool = True, page_params: dict[str, str] | None = None
    ) -> tuple[Sequence[Any], int]:
        count_query = select(func.count()).select_from(query.subquery())
        total = self.session.scalar(count_query)

        if total is None:
            total = 0  # TODO throw exception

        result_query = query.offset(page_params["offset"]).limit(page_params["limit"]) \
            if page_params else query
        items = self.session.scalars(result_query).all() if scalars else self.session.execute(result_query).all()

        return items, total

    def __filter(self, query: Select, filters: dict[str, Any]):
        conditions = [
            getattr(self.model, key) == value
            for key, value in filters.items()
        ]
        return query.where(*conditions)

    def __order_by(self, query: Select, order: list[str]):
        return query.order_by(*order)

    # TODO proper create and update method
    def create(self, obj_in: ModelType) -> ModelType:
        self.session.add(obj_in)
        self.session.commit()
        self.session.refresh(obj_in)
        return obj_in

    def update(self, obj_in: ModelType) -> ModelType:
        self.session.commit()
        self.session.refresh(obj_in)
        return obj_in

    def update_all(self, objs_in: list[ModelType]):
        self.session.commit()




import re
import datetime
import uuid
from sqlalchemy import TIMESTAMP, func
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Mapped, mapped_column, DeclarativeBase


# TODO https://alembic.sqlalchemy.org/en/latest/naming.html

#convention = {
#    "ix": "ix_%(column_0_label)s",
#    "uq": "uq_%(table_name)s_%(column_0_name)s",
#    "ck": "ck_%(table_name)s_%(constraint_name)s",
#    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
#    "pk": "pk_%(table_name)s",
#}

#metadata = MetaData(naming_convention=convention)

#Base = declarative_base(metadata=metadata)

class ModelBase(DeclarativeBase):

    __abstract__ = True

    @declared_attr.directive
    def __tablename__(self) -> str:
        names = re.split("(?=[A-Z])", self.__name__)
        return "_".join([x.lower() for x in names if x])


class UuidMixin:
    id: Mapped[uuid.UUID] = mapped_column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)


class AuditMixin:
    created_at: Mapped[datetime.datetime] = mapped_column(
        TIMESTAMP(timezone=True),
        default=func.now()
    )
    updated_at: Mapped[datetime.datetime] = mapped_column(
        TIMESTAMP(timezone=True),
        default=func.now(),
        onupdate=func.current_timestamp())




import typer
import psycopg


def setup_database(admin_config, db_name, db_user, db_password):
    conn_string = f"host={admin_config['host']} port={admin_config['port']} user={admin_config['user']} password={admin_config['password']}"

    with psycopg.connect(conn_string, autocommit=True) as conn:
        with conn.cursor() as cursor:
            cursor.execute("SELECT 1 FROM pg_database WHERE datname = %s", (db_name,))
            db_exists = cursor.fetchone()

            if not db_exists:
                typer.echo(f"Creating database '{db_name}'...")
                cursor.execute(f"CREATE DATABASE {db_name}")
                typer.echo(f"Database '{db_name}' created successfully!")
            else:
                typer.echo(f"Database '{db_name}' already exists.")

            cursor.execute("SELECT 1 FROM pg_roles WHERE rolname = %s", (db_user,))
            user_exists = cursor.fetchone()

            if not user_exists:
                typer.echo(f"Creating user '{db_user}'...")
                cursor.execute(f'CREATE USER "{db_user}" WITH PASSWORD \'{db_password}\'')
                typer.echo(f"User '{db_user}' created successfully!")
            else:
                typer.echo(f"User '{db_user}' already exists.")
                typer.echo(f"Updating password for user '{db_user}'...")
                cursor.execute("ALTER USER %s WITH PASSWORD %s", (db_user, db_password))

            typer.echo(f"Granting privileges on '{db_name}' to '{db_user}'...")
            cursor.execute(f"GRANT ALL PRIVILEGES ON DATABASE {db_name} TO {db_user}")

    db_conn_string = f"{conn_string} dbname={db_name}"

    with psycopg.connect(db_conn_string, autocommit=True) as conn:
        with conn.cursor() as cursor:
            cursor.execute(f"GRANT ALL ON SCHEMA public TO {db_user}")
            typer.echo(f"Schema privileges granted to '{db_user}'!")

    typer.echo("Database setup completed successfully!")



from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials
import secrets


def basic_auth(username: str = "admin", password: str = "admin"):
    def inner(
        credentials: HTTPBasicCredentials = Depends(HTTPBasic()),
    ) -> str:
        is_correct_username = secrets.compare_digest(credentials.username, username)
        is_correct_password = secrets.compare_digest(credentials.password, password)

        if not (is_correct_username and is_correct_password):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid authentication credentials",
                headers={"WWW-Authenticate": "Basic"},
            )

        return credentials.username

    return inner



from pydantic import PostgresDsn, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class NephyxBaseSettings(BaseSettings):

    model_config = SettingsConfigDict(
        env_prefix="APP_",
        env_file=".env",
    )

    DATABASE_HOST: str = "localhost"
    DATABASE_PORT: str = "5432"
    DATABASE_NAME: str
    DAYABASE_USER: str
    DATABASE_PASSWORD: str
    DATABASE_URL: PostgresDsn | None = None

    @field_validator("DATABASE_URL")
    @classmethod
    def assemble_database_url(cls, v: PostgresDsn | None, values):
        return PostgresDsn.build(
            scheme="postgresql+psycopg",
            username=values.data.get("database_user"),
            password=values.data.get("database_password"),
            host=values.data.get("database_host"),
            port=int(values.data.get("database_port")),
            path=values.data.get("database_name"),
        )



from typing import Generic, TypeVar
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class BaseSchema(BaseModel):
    model_config = ConfigDict(
        from_attributes=True,
        coerce_numbers_to_str=False,
        alias_generator=to_camel,
        populate_by_name=True,
    )


Data = TypeVar("Data", bound=(BaseSchema))


# TODO check if this is really required
class DataResponse(BaseSchema, Generic[Data]):
    success: bool = True
    data: Data | list[Data]


class SingleDataResponse(BaseSchema, Generic[Data]):
    success: bool = True
    data: Data


class ListDataResponse(BaseSchema, Generic[Data]):
    success: bool = True
    data: list[Data]


class PaginationMetadata(BaseSchema):
    page: int
    limit: int | None
    total: int
    pages: int


class PaginatedDataResponse(BaseSchema, Generic[Data]):
    success: bool = True
    data: list[Data]
    metadata: PaginationMetadata



import logging
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Callable, Dict, Optional

from fastapi import Depends
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)

logger = logging.getLogger(__name__)


class DatabaseManager:

    def __init__(self, connection_options: Dict[str, Any]):
        self.connection_options = connection_options
        self.engine: Optional[AsyncEngine] = None
        self.async_session_factory: Optional[async_sessionmaker[AsyncSession]] = None

    def init_engine(self) -> AsyncEngine:
        if self.engine is None:
            logger.info("Initializing database engine")
            self.engine = create_async_engine(
                self.connection_options.pop("url"),
                **self.connection_options
            )
            self.async_session_factory = async_sessionmaker(
                self.engine,
                expire_on_commit=False,
                autoflush=False,
            )

        return self.engine

    def get_session_factory(self) -> async_sessionmaker[AsyncSession]:
        if self.async_session_factory is None:
            self.init_engine()

        return self.async_session_factory

    @asynccontextmanager
    async def session(self) -> AsyncGenerator[AsyncSession, None]:
        session_factory = self.get_session_factory()

        async with session_factory() as session:
            try:
                yield session
                await session.commit()
            except Exception as e:
                await session.rollback()
                raise e

    async def close(self) -> None:
        if self.engine:
            logger.info("Closing database engine")
            await self.engine.dispose()
            self.engine = None
            self.async_session_factory = None


# Global database manager instance
db_manager: Optional[DatabaseManager] = None


def init_db(connection_options: Dict[str, Any]) -> DatabaseManager:
    global db_manager

    if db_manager is None:
        db_manager = DatabaseManager(connection_options)
        db_manager.init_engine()

    return db_manager


async def close_db() -> None:
    global db_manager

    if db_manager:
        await db_manager.close()
        db_manager = None


async def get_db_session() -> AsyncGenerator[AsyncSession, None]:
    if db_manager is None:
        raise RuntimeError("Database not initialized. Call init_db first.")

    async with db_manager.session() as session:
        yield session


# Dependency for injecting DB session
DbSession = Callable[[], AsyncGenerator[AsyncSession, None]]
db_session_dependency = Depends(get_db_session)
